# File tres/Src/opspec.x
# March 6, 2009
# By Doug Mink, Harvard-Smithsonian Center for Astrophysics

# Copyright(c) 2009 Smithsonian Astrophysical Observatory
# You may do anything you like with this file except remove this copyright.
# The Smithsonian Astrophysical Observatory makes no representations about
# the suitability of this software for any purpose.  It is provided "as is"
# without express or implied warranty.
 
# OPSPEC rebins an order of a second spectrum to match a first spectrum and then
#        performs an operation on them
 
include	<imhdr.h>
include	<imio.h>
include	<fset.h>
include <smw.h>

# operation defs
define OP_ADD		1
define OP_SUB		2
define OP_MUL		3
define OP_DIV		4
define OP_FUNCTIONS	"|+|-|*|/|"

procedure opspec (spec1,spath1,spec2,spath2,order,norders,op,imres,respath,resname,resspec)

char	spec1[ARB]	# First spectrum file name
char	spath1[ARB]	# Path to first spectrum
char	spec2[ARB]	# Second spectrum file name
char	spath2[ARB]	# Path to second spectrum
int	order		# Object aperture to read from multispec file
int	norders		# Number of input apertures to read in total
int	op		# Operation (+-*/ -> 1234)
pointer imres		# Composite image header structure
char	respath[ARB]	# Composite spectrum file path
char	resname[ARB]	# Title for output spectrum
pointer resspec		# Composite spectrum

char	sname1[SZ_IMTITLE]	# Title for first spectrum
pointer spix1		# First spectrum
pointer spim1		# First spectrum image header structure
pointer spsh1		# First spectrum header structure
char	sname2[SZ_IMTITLE]	# Title for second spectrum
pointer spix2		# Second spectrum
pointer spim2		# Second spectrum image header structure
pointer spsh2		# Second spectrum header structure
int	npix1, npix2
int	i, j
char	wtitle[20]	# Wavelength axis caption for plots of spectrum
char	title[SZ_LINE]	# Title for plots of spectrum
char	tstr[SZ_LINE]

char	fstring[8]
char	opstr[4]

double	px
real	rindef
bool	pixfill
int	npwcs1, npwcs2
int	nsig

pointer	wlwcs1, wlwcs2
pointer	spwcs1, spwcs2, spwcs2r, reswcsr
int	mode

int	clscan()
#int	imaccess()
double	shdr_lw()
int	strcmp()
errchk	malloc

define	endspec_	 90

include	"tresop.com"

 
begin
	spsh1 = NULL
	spsh2 = NULL
	call sprintf (wtitle,20,"Wavelength")
	nsig = -1
	mode = READ_ONLY
	rindef = INDEFR

# Load first spectrum
	call gettres (spath1, order, spix1, spim1, spsh1, norders,
		      npix1, sname1, mode, debug)
	if (spim1 == ERR) {
	    call eprintf ("OPSPEC: Error reading spectrum %s\n")
	 	call pargstr (spath2)
	    go to endspec_
	    }

# Compute the wavelength vector for the first spectrum.
	if (npix1 > npix01) {
	    if (wl1 != NULL)
		call mfree (wl1, TY_REAL)
	    call malloc (wl1,npix1,TY_REAL)
	    }
	do j = 1, npix1 {
	    px = double (j)
	    if (j < NP1(spsh1))
		Memr[wl1+j-1] = 0.0
	    else if (j > NP2(spsh1))
		Memr[wl1+j-1] = 0.0
	    else
		Memr[wl1+j-1] = real (shdr_lw (spsh1, px))
	    }

# Set pointers to wavelength-calibrated data
	npwcs1 = NP2(spsh1) - NP1(spsh1) + 1
	spwcs1 = spix1 + NP1(spsh1) - 1
	wlwcs1 = wl1 + NP1(spsh1) - 1
	if (debug) {
	    call printf ("OPSPEC: %d points in file %s[%d]")
		call pargi (npwcs1)
		call pargstr (spec1)
		call pargi (order)
	    if (ymin != rindef && ymax != rindef) {
		call printf (" from %f-%f\n")
		    call pargr (ymin)
		    call pargr (ymax)
		}
	    else
		call printf ("\n")
	    call flush (STDOUT)
	    }

# If plot enabled, show the first spectrum.
	if (pltspec) {
	    call sprintf (title,SZ_LINE,"%s[%d]")
		call pargstr (spec1)
		call pargi (order)
	    call tresplot (npwcs1, Memr[spwcs1], title, Memr[wlwcs1],
			   wtitle, ssmooth, intspec, ymin, ymax)
	    }

# Eliminate bad lines from first spectrum
	if (clscan("fixbad") != EOF)
	    call clgstr ("fixbad", fstring, 8)
	else
	    call strcpy ("no", fstring, 8)
	if (strcmp (fstring, "no") != 0 ) {
	    call filllist (npwcs1, Memr[spwcs1], Memr[wlwcs1], order, pixfill, debug)
	    if (pltspec) {
		call sprintf (title,SZ_LINE,"%s[%d] filling bad lines")
		    call pargstr (spec1)
		    call pargi (order)
		call tresplot (npwcs1, Memr[spwcs1], title, Memr[wlwcs1],
			       wtitle, ssmooth, intspec, ymin, ymax)
		}
	    }

# Load second spectrum
	call gettres (spath2, order, spix2, spim2, spsh2, norders,
		      npix2, sname2, mode, debug)
	if (spim2 == ERR) {
	    call eprintf ("OPSPEC: Error reading spectrum %s\n")
	 	call pargstr (spath2)
	    go to endspec_
	    }

# Compute the wavelength vector for the second spectrum.
	if (npix2 > npix02) {
	    if (wl2 != NULL)
		call mfree (wl2, TY_REAL)
	    call malloc (wl2,npix2,TY_REAL)
	    npix02 = npix2
	    }
	do j = 1, npix2 {
	    px = double (j)
	    if (j < NP1(spsh2))
		Memr[wl2+j-1] = 0.0
	    else if (j > NP2(spsh2))
		Memr[wl2+j-1] = 0.0
	    else
		Memr[wl2+j-1] = real (shdr_lw (spsh2, px))
	    }

# Set pointers to wavelength-calibrated data
	npwcs2 = NP2(spsh2) - NP1(spsh2) + 1
	spwcs2 = spix2 + NP1(spsh2) - 1
	wlwcs2 = wl2 + NP1(spsh2) - 1
	if (debug) {
	    call printf ("OPSPEC: %d points in file %s[%d]")
		call pargi (npwcs2)
		call pargstr (spec2)
		call pargi (order)
	    if (ymin != rindef && ymax != rindef) {
		call printf (" from %f-%f\n")
		    call pargr (ymin)
		    call pargr (ymax)
		}
	    else
		call printf ("\n")
	    call flush (STDOUT)
	    }

# If plot enabled, show the second spectrum.
	if (pltspec) {
	    call sprintf (title,SZ_LINE,"%s[%d]")
		call pargstr (spec2)
		call pargi (order)
	    call tresplot (npwcs2, Memr[spwcs2], title, Memr[wlwcs2],
			   wtitle, ssmooth, intspec, ymin, ymax)
	    }

# Eliminate bad lines from second spectrum
	if (strcmp (fstring, "no") != 0 ) {
	    call filllist (npwcs2, Memr[spwcs2], Memr[wlwcs2], order, pixfill, debug)
	    if (pltspec) {
		call sprintf (title,SZ_LINE,"%s[%d] filling bad lines")
		    call pargstr (spec2)
		    call pargi (order)
		call tresplot (npwcs2,Memr[spwcs2],title, Memr[wlwcs2],
			       wtitle, ssmooth, intspec, ymin, ymax)
		}
	    }

# Allocate rebinned spectra and wavelength vectors
	if (npix1 > npix01) {
	    if (spix2r != NULL)
		call mfree (spix2r, TY_REAL)
	    iferr (call malloc (spix2r, npix1, TY_REAL)) {
		call fprintf ("OPSPEC: Cannot allocate %d pixel rebin vector\n")
		    call pargi (npix1)
		call flush (STDOUT)
		return
		}
	    npix01 = npix1
	    if (debug) {
		call printf ("OPSPEC: %d pixel rebin vector allocated\n")
		    call pargi (npix1)
		call flush (STDOUT)
		}
	    }

#  Rebin this order of the second spectrum to match first spectrum
	call rebinm (spsh2, spsh1, Memr[spix2], Memr[spix2r], debug)
	spwcs2r = spix2r + NP1(spsh1) - 1
	if (debug) {
	    call printf ("OPSPEC: %d points in rebinned file %s[%d] from %d")
		call pargi (npwcs1)
		call pargstr (spec2)
		call pargi (order)
		call pargi (NP1(spsh1))
	    if (ymin != rindef && ymax != rindef) {
		call printf (" from %f-%f\n")
		    call pargr (ymin)
		    call pargr (ymax)
		}
	    else
		call printf ("\n")
	    call flush (STDOUT)
	    }

# If plot enabled, plot the rebinned second spectrum.
	if (pltspec) {
	    call sprintf (title,SZ_LINE,"%s[%d] rebinned")
		call pargstr (spec2)
		call pargi (order)
	    call tresplot (npwcs1,Memr[spwcs2r], title, Memr[wlwcs1],
			   wtitle, ssmooth, intspec, ymin, ymax)
	    }

# Perform vector operation using rebinned data
	if (op == OP_ADD) {
	    do i = 0, npix1-1 {
		Memr[resspec+i] = Memr[spix1+i] + Memr[spix2r+i]
		}
	    call strcpy ("+", opstr, 4)
	    }
	else if (op == OP_SUB) {
	    do i = 0, npix1-1 {
		Memr[resspec+i] = Memr[spix1+i] - Memr[spix2r+i]
		}
	    call strcpy ("-", opstr, 4)
	    }
	else if (op == OP_MUL) {
	    do i = 0, npix1-1 {
		Memr[resspec+i] = Memr[spix1+i] * Memr[spix2r+i]
		}
	    call strcpy ("*", opstr, 4)
	    }
	else if (op == OP_DIV) {
	    do i = 0, npix1-1 {

#	    If pixel is zero, use the value of the closest + 1 non-zero pixel
		if (Memr[spix2r+i] == 0.0 || Memr[spix1+i] == 0.0) {
		    if (i < npix1 / 2) {
			j = i
			while (Memr[spix2r+j] == 0.0 || Memr[spix1+j] == 0.0)
			    j = j + 1
			Memr[resspec+i] = Memr[spix1+j+1] / Memr[spix2r+j+1]
			}
		    else {
			j = i
			while (Memr[spix2r+j] == 0.0 || Memr[spix1+j] == 0.0)
			     j = j - 1
			Memr[resspec+i] = Memr[spix1+j-1] / Memr[spix2r+j-1]
			}
		    }
		else if (i > 1 && Memr[spix2r+i-1] == 0.0) {
		    Memr[resspec+i] = Memr[spix1+i+1] / Memr[spix2r+i+1]
		    }
		else if (i < npix1-1 && Memr[spix2r+i+1] == 0.0) {
		    Memr[resspec+i] = Memr[spix1+i-1] / Memr[spix2r+i-1]
		    }
		else {
		    Memr[resspec+i] = Memr[spix1+i] / Memr[spix2r+i]
		    if (debug && i < 30) {
			call printf ("OPSPEC: %8g %7g / %7g = %7g\n")
			    call pargr (Memr[wlwcs1+i])
			    call pargr (Memr[spix1+i])
			    call pargr (Memr[spix2r+i])
			    call pargr (Memr[resspec+i])
			}
		    }
		}
	    call strcpy ("/", opstr, 4)
	    }
	else {
	    do i = 0, npix1-1 {
		Memr[resspec+i] = Memr[spix2r+i]
		}
	    call strcpy ("bin", opstr, 4)
	    }

# Smooth result
	reswcsr = resspec + NP1(spsh1) - 1
	if (rsmooth > 0)
	    call smooth (Memr[reswcsr],npix1,rsmooth)

# Note operation for this order
	call sprintf (title,SZ_LINE,"%s[%d] %s %s[%d]")
	    call pargstr (spec1)
	    call pargi (order)
	    call pargstr (opstr)
	    call pargstr (spec2)
	    call pargi (order)
	if (rsmooth > 0) {
	    call sprintf (tstr, SZ_LINE, " smoothed %d")
		call pargi (rsmooth)
	    call strcat (tstr, title, SZ_LINE)
	    }
	call imastr (imres, "HISTORY", title)

# Check for writability
#	if (imaccess (respath, WRITE_ONLY) == NO) {
#	    call eprintf ("OPSPEC: cannot write to %s; not saving results\n")
#		call pargstr (respath)
#	    IM_UPDATE(imres) = NO
#	    }
#	else
	    IM_UPDATE(imres) = YES

# If plot enabled, plot the result spectrum
	if (pltfit) {
	    call sprintf (title,SZ_LINE,"%s[%d] of %d")
		call pargstr (resname)
		call pargi (order)
		call pargi (norders)
	    call tresplot (npix1,Memr[reswcsr],title, Memr[wlwcs1],
			   wtitle, 0, intfit, ymin,ymax)
	    }

	call fitspec (npix1, Memr[reswcsr], Memr[wlwcs1], title, ssmooth)

# If plot enabled, plot the result spectrum
	if (pltres) {
	    call sprintf (title,SZ_LINE,"%s[%d] of %d")
		call pargstr (resname)
		call pargi (order)
		call pargi (norders)
	    call tresplot (npix1,Memr[reswcsr],title, Memr[wlwcs1],
			   wtitle, 0, intres, ymin,ymax)
	    }

# Write the resulting spectrum order
	call res_write (imres, order, resspec, TY_REAL, debug)

# Free rebinned spectra, wavelength, and correlation vectors
endspec_

# Close the input spectrum images
	if (debug) {
	    call printf ("OPSPEC: About to close %s\n")
		call pargstr (spec1)
	    call flush (STDOUT)
	    }
	IM_UPDATE(spim1) = NO
	call closetres (spim1, spsh1)
	if (debug) {
	    call printf ("OPSPEC: About to close %s\n")
		call pargstr (spec2)
	    call flush (STDOUT)
	    }
	IM_UPDATE(spim2) = NO
	call closetres (spim2, spsh2)
	if (debug) {
	    call printf ("OPSPEC: Input files %s[%d] and %s[%d] closed\n")
		call pargstr (spec1)
		call pargi (order)
		call pargstr (spec2)
		call pargi (order)
	    call flush (STDOUT)
	    }
	return

end

# Feb  9 2009	New program based on rvsao/Sumspec/addspec.x
# Mar  6 2009	Drop unused variables
